using System;
using System.Text;

namespace FrameworkQ.ObjectSpace.Data.SqlObjects.Generation
{
	/// <summary>
	/// Renderer for Oracle
	/// </summary>
	/// <remarks>
	/// Use OracleRenderer to render SQL statements for Oracle database.
	/// This version of Sql.Net has been tested with Oracle 9i.
	/// </remarks>
	public class OracleGenerator : SqlGenerator
	{
		/// <summary>
		/// Creates a new instance of OracleRenderer
		/// </summary>
		public OracleGenerator() : base('"', '"')
		{
			DateFormat = "dd-MMM-yy";
			DateTimeFormat = "dd-MMM-yy HH:mm:ss";
		}

		/// <summary>
		/// Renders IfNull SqlExpression
		/// </summary>
		/// <param name="builder"></param>
		/// <param name="expr"></param>
		protected override void IfNull(StringBuilder builder, SqlExpression expr)
		{
			builder.Append("nvl(");
			Expression(builder, expr.SubExpr1);
			builder.Append(", ");
			Expression(builder, expr.SubExpr2);
			builder.Append(")");
		}

		/// <summary>
		/// Returns true. 
		/// </summary>
		protected override bool UpperCaseIdentifiers
		{
			get { return true; }
		}

		/// <summary>
		/// Renders bitwise and
		/// </summary>
		/// <param name="builder"></param>
		/// <param name="item"></param>
		protected override void BitwiseAnd(StringBuilder builder, WhereItem item)
		{
			builder.Append("BITAND(");
			Expression(builder, item.Expr1);
			builder.Append(", ");
			Expression(builder, item.Expr2);
			builder.Append(") > 0");
		}

		/// <summary>
		/// Renders a SELECT statement
		/// </summary>
		/// <param name="query">Query definition</param>
		/// <returns>Generated SQL statement</returns>
		public override string GenerateSelectStatement(SelectQuery query)
		{
			if (query.Top > -1 && query.OrderByItems.Count > 0)
			{
				string baseSql = RenderSelect(query, -1);

				SelectQuery countQuery = new SelectQuery();
				SelectColumn col = new SelectColumn("*");
				countQuery.Columns.Add(col);
				countQuery.FromClause.BaseTable = FromItem.SubQuery(baseSql, "t");
				return RenderSelect(countQuery, query.Top);
			}
			else
				return RenderSelect(query, query.Top);
		}

		string RenderSelect(SelectQuery query, int limitRows)
		{
			query.Validate();
			
			StringBuilder selectBuilder = new StringBuilder();

			//Start the select statement
			this.Select(selectBuilder, query.Distinct);
			
			//Render select columns
			this.SelectColumns(selectBuilder, query.Columns);

			this.FromClause(selectBuilder, query.FromClause, query.TableSpace);

			WhereClause fullWhereClause = new WhereClause(WhereClauseRelationship.And);
			fullWhereClause.SubClauses.Add(query.WherePhrase);
			if (limitRows > -1)
				fullWhereClause.Items.Add(WhereItem.CreateCompare(SqlExpression.PseudoField("rownum"), SqlExpression.Number(limitRows), CompareOperator.LessOrEqual));

			this.Where(selectBuilder, fullWhereClause);
			this.WhereClause(selectBuilder, fullWhereClause);

			this.GroupBy(selectBuilder, query.GroupByItems);
			if (query.GroupByWithCube)
				selectBuilder.Append(" cube (");
			else if (query.GroupByWithRollup)
				selectBuilder.Append(" rollup (");
			this.GroupByTerms(selectBuilder, query.GroupByItems);

			if (query.GroupByWithCube || query.GroupByWithRollup)
				selectBuilder.Append(" )");
			
			this.Having(selectBuilder, query.HavingPhrase) ;
			this.WhereClause(selectBuilder, query.HavingPhrase);

			this.OrderBy(selectBuilder, query.OrderByItems);
			this.OrderByTerms(selectBuilder, query.OrderByItems);

			return selectBuilder.ToString();
		}

		/// <summary>
		/// Renders a row count SELECT statement. 
		/// </summary>
		/// <param name="query">Query definition to count rows for</param>
		/// <returns>Generated SQL statement</returns>
		/// <remarks>
		/// Renders a SQL statement which returns a result set with one row and one cell which contains the number of rows <paramref name="query"/> can generate. 
		/// The generated statement will work nicely with <see cref="System.Data.IDbCommand.ExecuteScalar"/> method.
		/// </remarks>
		public override string GenerateRowCountStatement(SelectQuery query)
		{
			string baseSql = RenderSelect(query, -1);

			SelectQuery countQuery = new SelectQuery();
			SelectColumn col = new SelectColumn("*", null, "cnt", SqlAggregationFunction.Count);
			countQuery.Columns.Add(col);
			countQuery.FromClause.BaseTable = FromItem.SubQuery(baseSql, "t");
			return GenerateSelectStatement(countQuery);
		}
	}
}
