using System;
using System.Collections;

namespace FrameworkQ.ObjectSpace.Data.SqlObjects
{
	internal enum WhereItemType { Compare, Between, In, NotIn, InSubQuery, NotInSubQuery, IsNull, IsNotNull, Exists, NotExists }
	
	/// <summary>
	/// Represents one term in a WHERE clause
	/// </summary>
	/// <remarks>
	/// <see cref="WhereItem"/> usually consists of one or more <see cref="SqlExpression"/> objects and an a conditional operator which applies to those expressions.
	/// <see cref="WhereItem"/> has no public constructor. Use one of the supplied static methods to create a term. 
	/// <para>
	/// Use <see cref="WhereItem.CreateCompare"/> to create a comparison term. A comparison term can apply one of <see cref="CompareOperator"/> operators on the supplied expressions.
	/// Use <see cref="WhereItem.CreateIn"/> to create a term which checks wheather an expression exists in a list of supplied values.
	/// Use <see cref="WhereItem.CreateBetween"/> to create a term which checks wheather an expression value is between a supplied lower and upper bounds.
	/// </para>
	/// </remarks>
	public class WhereItem : ICloneable
	{
		SqlExpression expr1, expr2, expr3;
		CompareOperator op;
		WhereItemType type;
		SqlConstantCollection values;
		string subQuery;

		private WhereItem()
		{
		}

		/// <summary>
		/// Creates a comparison WhereTerm.
		/// </summary>
		/// <param name="expr1">Expression on the left side of the operator</param>
		/// <param name="expr2">Expression on the right side of the operator</param>
		/// <param name="op">Conditional operator to be applied on the expressions</param>
		/// <returns>A new conditional WhereTerm</returns>
		/// <remarks>
		/// A comparison term compares two expression on the basis of their values. Expressions can be of any type but their results must be of comparible types. 
		/// For instance, you can not compare a database field of type 'date' and a static value of type 'int'.
		/// </remarks>
		/// <example>
		/// <code>
		/// ...
		/// query.WherePhrase.Terms.Add(WhereTerm.CreateCompare(SqlExpression.Field("name", tCustomers), SqlExpression.String("J%"), CompareOperator.Like));
		/// </code>
		/// </example>
		public static WhereItem CreateCompare(SqlExpression expr1, SqlExpression expr2, CompareOperator op)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr1;
			item.expr2 = expr2;
			item.op = op;
			
			item.type = WhereItemType.Compare;
			
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which represents SQL IN clause
		/// </summary>
		/// <param name="expr">Expression to be looked up</param>
		/// <param name="sql">Sub query</param>
		/// <returns></returns>
		public static WhereItem CreateIn(SqlExpression expr, string sql)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr;
			item.subQuery = sql;

			item.type = WhereItemType.InSubQuery;
			
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which represents SQL IN clause
		/// </summary>
		/// <param name="expr">Expression to be looked up</param>
		/// <param name="values">List of values</param>
		/// <returns></returns>
		public static WhereItem CreateIn(SqlExpression expr, SqlConstantCollection values)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr;
			item.values = values;

			item.type = WhereItemType.In;
			
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which represents SQL NOT IN clause
		/// </summary>
		/// <param name="expr">Expression to be looked up</param>
		/// <param name="sql">Sub query</param>
		/// <returns></returns>
		public static WhereItem CreateNotIn(SqlExpression expr, string sql)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr;
			item.subQuery = sql;

			item.type = WhereItemType.NotInSubQuery;
			
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which represents SQL NOT IN clause
		/// </summary>
		/// <param name="expr">Expression to be looked up</param>
		/// <param name="values"></param>
		/// <returns></returns>
		public static WhereItem CreateNotIn(SqlExpression expr, SqlConstantCollection values)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr;
			item.values = values;

			item.type = WhereItemType.NotIn;
			
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which returns TRUE if an expression is NULL
		/// </summary>
		/// <param name="expr">Expression to be evaluated</param>
		/// <returns></returns>
		public static WhereItem CreateIsNull(SqlExpression expr)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr;
			item.type = WhereItemType.IsNull;
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which returns TRUE if an expression is NOT NULL
		/// </summary>
		/// <param name="expr"></param>
		/// <returns></returns>
		public static WhereItem CreateIsNotNull(SqlExpression expr)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr;
			item.type = WhereItemType.IsNotNull;
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which encapsulates SQL EXISTS clause
		/// </summary>
		/// <param name="sql">Sub query for the EXISTS clause</param>
		/// <returns></returns>
		public static WhereItem CreateExists(string sql)
		{
			WhereItem item = new WhereItem();
			item.subQuery = sql;
			item.type = WhereItemType.Exists;
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which encapsulates SQL NOT EXISTS clause
		/// </summary>
		/// <param name="sql">Sub query for the NOT EXISTS clause</param>
		/// <returns></returns>
		public static WhereItem CreateNotExists(string sql)
		{
			WhereItem item = new WhereItem();
			item.subQuery = sql;
			item.type = WhereItemType.NotExists;
			return item;
		}

		/// <summary>
		/// Creates a WhereTerm which checks weather a value is in a specifed range.
		/// </summary>
		/// <param name="expr">Expression which yeilds the value to be checked</param>
		/// <param name="lowBound">Expression which yeilds the low bound of the range</param>
		/// <param name="highBound">Expression which yeilds the high bound of the range</param>
		/// <returns>A new WhereTerm</returns>
		/// <remarks>
		/// CreateBetween only accepts expressions which yeild a 'Date' or 'Number' values.
		/// All expressions must be of compatible types.
		/// </remarks>
		public static WhereItem CreateBetween(SqlExpression expr, SqlExpression lowBound, SqlExpression highBound)
		{
			WhereItem item = new WhereItem();
			item.expr1 = expr;
			item.expr2 = lowBound;
			item.expr3 = highBound;
			
			item.type = WhereItemType.Between;			
			
			return item;
		}
		
		internal SqlExpression Expr1
		{
			get { return expr1; }
		}
		
		internal SqlExpression Expr2
		{
			get { return expr2; }
		}
		
		internal SqlExpression Expr3
		{
			get { return expr3;	}
		}
		
		internal CompareOperator Op
		{
			get { return op; }
		}
		
		internal WhereItemType Type
		{
			get { return type; }
		}
		
		internal SqlConstantCollection Values
		{
			get { return values; }
		}

		internal string SubQuery
		{
			get { return this.subQuery; }
		}
		
		object ICloneable.Clone()
		{
			return this.Clone();
		}

		/// <summary>
		/// Creates a copy of this WhereTerm
		/// </summary>
		/// <returns>A new WhereTerm which exactly the same as the current one.</returns>
		public WhereItem Clone()
		{
			WhereItem a = new WhereItem();

			a.expr1 = expr1;
			a.expr2 = expr2;
			a.expr3 = expr3;
			a.op = op;
			a.type = type;
			a.subQuery = a.subQuery;
			a.values = new SqlConstantCollection(values);
			
			return a;
		}
	}
}
