//	Enterprise.ObjectSpace.Database
//	Copyright (C) 2004 by Shafqat Ahmed
//  ===============================================================================
//  Contact Information
//	Email   : admin@shiplu.com
//	IM      : raasiel@hotmail.com
//  -------------------------------------------------------------------------------
//	Shared Source License for FrameworkQ
//  -------------------------------------------------------------------------------
//	This license governs use of the accompanying software ('FrameworkQ.ObjectSpace'), 
//  and your use of the Software constitutes acceptance of this license.
//
//	You may use the Software for any commercial or noncommercial purpose,
//	including distributing derivative works. The derivative works must contain
//	acknowledgement to the original work.
//
//	In return, we simply require that you agree:
//  -------------------------------------------------------------------------------
//	1. Not to remove any copyright or other notices from the Software. 
//
//	2. US jurisdiction does not apply on this product since the author is not 
//	   a US citizen and the product was made outside US. However any extension 
//	   of the project may fall under US jurisdiction if the extender developer
//	   is a US citizen or is located at US.
//
//	3. That if you distribute the Software in source code form you do so only
//	   under this license, and if you distribute the Software solely in
//	   object form you only do so under a license that complies with this
//	   license.
//
//	4. That the Software comes "as is", with no warranties. None whatsoever.
//	   This means no express, implied or statutory warranty, including without
//	   limitation, warranties of merchantability or fitness for a particular
//	   purpose or any warranty of title or non-infringement. Also, you must pass
//	   this disclaimer on whenever you distribute the Software or derivative
//	   works.
//
//	5. That no contributor to the Software will be liable for any of those types
//	   of damages known as indirect, special, consequential, or incidental
//	   related to the Software or this license, to the maximum extent the law
//	   permits, no matter what legal theory its based on. Also, you must pass
//	   this limitation of liability on whenever you distribute the Software or
//	   derivative works.
//
//	6. That if you sue anyone over patents that you think may apply to the
//	   Software for a person's use of the Software, your license to the Software
//	   ends automatically.
//
//	7. That the patent rights, if any, granted in this license only apply to the
//	   Software, not to any derivative works you make.
//
//	8. That your rights under this License end automatically if you breach it in
//	   any way.
//
//	9. That all rights not expressly granted to you in this license are reserved.
//	   executable file might be covered by the GNU General Public License.
//  ==============================================================================
//  Visit FrameworkQ.ObjectSpace workspace at :
//  http://www.gotdotnet.com/Community/Workspaces/Workspace.aspx?id=d6298665-58a2-4e33-b7f7-417031ac9a64


using System;
using System.Collections;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Diagnostics;
using System.Reflection;
using System.Security.Principal;
using FrameworkQ.ObjectSpace.Data.Schema;
using FrameworkQ.ObjectSpace.Util;
using FrameworkQ.ObjectSpace.Data.Providers;
using FrameworkQ.ObjectSpace.Data.SqlObjects;

namespace FrameworkQ.ObjectSpace
{
	

	/// <summary>
	/// Database is the core class for the solution. This class contains the 
	/// code objectspace functionality.
	/// 
	/// Please rememeber that, since the database caches the type definitions
	/// to the table schemas, in order to avoid gettting schema for each it
	/// is advised to store the database instance.
	/// </summary>
	public class Database :IDisposable
	{
		public const string CONNECTION_STRING_KEY = "FrameworkQ.ObjectSpace.ConnectionString";
		public const string PROVIDER_KEY = "FrameworkQ.ObjectSpace.Provider";

		private DatabaseGraph _graph =null;
		private bool _pooled = false;
		
		private Configuration _configuration = null;
	
		/// <summary>
		/// Constructs the database with a conenction string. 
		/// </summary>
		/// <param name="connectionString"></param>
		/// <example>
		///		The following example creates a database object
		///		<code>
		///			string connectionString ="Server=localhost;Database=ObjectSpace;Trusted_Connection=True;";
		///			// Creates a new database
		///			Database database = new Database(connectionString);
		///		</code>
		/// </example>
		public Database(Configuration configuration, bool pooled)
		{
			_configuration = configuration;
			_pooled = pooled;
			DatabaseGraph databaseGraph = this.GetGraph();
		}		


		/// <summary>
		/// Constructs the database.
		/// </summary>
		/// <remarks>
		///		There must be a connection string in the config file under this key in the
		///		the appsettings section 
		///			"FrameworkQ.ObjectSpace.ConnectionString"
		/// </remarks>
		/// <example>
		///		The following example creates a database object
		///		<code>
		///			string connectionString ="Server=localhost;Database=ObjectSpace;Trusted_Connection=True;";
		///			// Creates a new database
		///			Database database = new Database();
		///		</code>
		/// </example>
		public Database(bool pooled)
		{
			/// Since no one assigend any connection string lets try to find one ...			
			string connectionString = ConfigurationManager.AppSettings[Database.CONNECTION_STRING_KEY];
			string providerClass = ConfigurationManager.AppSettings[Database.PROVIDER_KEY];

			_pooled = pooled;
			if (connectionString!=null && providerClass!=null)
			{
				if (connectionString.Length>0 && providerClass.Length>0)
				{
					Type type = Type.GetType (providerClass,true);
					IDatabaseProvider provider = Activator.CreateInstance(type) as IDatabaseProvider;
					_configuration = new Configuration(provider, connectionString);
					DatabaseGraph databaseGraph = this.GetGraph();					
				}
				else
				{
					throw new InvalidOperationException ("Must have the connection string and provider class" +
						" defined in configuration appsettings sections under the key " + 
						"\"FrameworkQ.ObjectSpace.ConnectionString\" and \"FrameworkQ.ObjectSpace.Provider\"" + 
						" when using parameterless constructor.");
				}
			}
			else
			{
				throw new InvalidOperationException("Must have the connection string and provider class" +
					" defined in configuration appsettings sections under the key " +
					"\"FrameworkQ.ObjectSpace.ConnectionString\" and \"FrameworkQ.ObjectSpace.Provider\"" +
					" when using parameterless constructor.");
			}
		}


		public Database ():this(true)
		{
		}

		/// <summary>
		/// Starts a SQL transaction
		/// </summary>
		/// <param name="level">The isolation level for the transaction</param>
		/// <returns>The created transaction</returns>
		/// <example>
		///		<code>
		///			string connectionString ="Server=localhost;Database=ObjectSpace;Trusted_Connection=True;";
		///			// Creates a new database
		///			Database database = new Database(connectionString);
		///			database.ConnectionKeepAlive = true;
		///			SqlTransaction transaction  = database.StartTransaction (IsolationLevel.ReadCommitted);
		///			database.Insert (someObject);
		///			transaction.Commit();
		///			database.EndTransaction();
		///		</code>
		/// </example>
		/// <remarks>
		///		While a transaction is being used. The KeepAlive sertting is overridden and
		///		the connection is kept alive.
		/// </remarks>
		public IDbTransaction BeginTransaction(IsolationLevel level)		
		{			
			//throw new NotImplementedException ();			
			IDbTransaction tran =  this.GetGraph().Connection.BeginTransaction (level);
			return tran;
		}


		/// <summary>
		/// Ends a transaction
		/// </summary>
		public void EndTransaction()
		{
			//throw new NotImplementedException ();
			this.GetGraph().Connection.EndTransaction();
		}
		

		
		#region "Data Modification Functions"

		/// <summary>
		/// Inserts and object to the database
		/// </summary>
		/// <param name="obj">Object to insert</param>
		/// <returns>
		/// 1.Returns the primary key if it is identity 
		///	2.Returns the rowcount if it is not and identity	
		/// </returns>
		/// <example>
		///		Suppose we have a user object that looks like this.
		///		<code>
		///	class User
		///	{
		///		public User(){}
		///
		///		private int _id=0;
		///		// Mark this as primary key for object space to identify
		///		[FrameworkQ.ObjectSpace.Schema.PrimaryKey]
		///		public int ID
		///		{
		///			get{return _id;}
		///			set{_id=value;}
		///		}
		///
		///		private string _name="";
		///		public string Name
		///		{
		///			get{return _name;}
		///			set{_name=value;}	
		///		}
		///
		///		private string _password="";
		///		public string Password
		///		{
		///			get{return _password;}
		///			set{_password=value;}	
		///		}
		///	}
		///		</code>
		///		In order to insert a new user object to database we can write the following code
		///		<code>
		///			
		///			string connectionString ="Server=localhost;Database=ObjectSpace;Trusted_Connection=True;";
		///			// Creates a new database
		///			Database database = new Database(connectionString);
		///			User user = new User();
		///			user.Name = "Someuser";
		///			user.Password = "123";
		///			// Insert the user object to the database
		///			database.Insert (user);
		///			// Notice that the ID property has been updated with the 
		///			// identity columns value for the object in database.
		///		</code>
		/// </example>
		/// <value></value>
		public int Insert (object obj) 
		{
            ObjectDataAccess data = GetObjectDataAccess(obj.GetType());
			return data.Insert (obj);
		}

		/// <summary>
		/// Updates an object
		/// </summary>
		/// <param name="obj">Object to update</param>
		public int Update (object obj) 
		{
			ObjectDataAccess data = GetObjectDataAccess (obj.GetType());
			return ( int )data.Update (obj);
		}

		/// <summary>
		/// Deletes an object from database
		/// </summary>
		/// <param name="objectType">Type information of the object to delete</param>
		/// <param name="primary">Primary key to the object</param>
		public int Delete(Type objectType, object primary) 
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			return data.Delete (primary);
		}

        public int DeleteBy(Type objectType, WhereClause clause)
        {
            ObjectDataAccess data = GetObjectDataAccess(objectType);
            return data.DeleteBy(clause);
        }

		/// <summary>
		/// Deletes all object of the specified type
		/// </summary>
		/// <param name="objectType">Type information of the object to delete</param>
		public void DeleteAll( Type objectType) 
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			data.DeleteAll ();
		}
		
		#endregion 

		#region "Data Selection Functions"
		/// <summary>
		/// Retrieves an object from the databse
		/// </summary>
		/// <param name="objectType">Type of object</param>
		/// <param name="primary">The primary key to the object</param>
		/// <returns>The object is question</returns>
		/// <example>
		///		Suppose we have a user object. See the example in <see cref="FrameworkQ.ObjectSpace.Insert"/> Insert.
		///		Now to retrieve such an object ...
		///		<code>
		///			string connectionString ="Server=localhost;Database=ObjectSpace;Trusted_Connection=True;";
		///			// Creates a new database
		///			Database database = new Database(connectionString);
		///			// Retrieve the user obejct with ID 1
		///			User user = (User) database.Select (typeof(User),1)
		///		</code>
		/// </example>
        public object Select (Type objectType, object primary)
        {
            ObjectDataAccess data = GetObjectDataAccess(objectType);
            return data.Select(primary);
        }

		public T Select <T>(object primary) 
		{
            return (T)this.Select(typeof(T), primary);
		}


		/// <summary>
		/// Gets all the objects of a certain type form the database
		/// </summary>
		/// <param name="objectType">Type information of the object to select</param>
		/// <returns>An IList of all the objects in database</returns>
		public IList SelectAll (Type objectType) 
		{
			return SelectAll (objectType, typeof (ArrayList));
		}

        public List<T> SelectAll<T>()
        {
            return (List<T>) SelectAll (typeof(T), typeof(List<T>));
        }



		/// <summary>
		/// Gets all the objects of a certain type form the database
		/// </summary>
		/// <param name="objectType">Type information of the object to select</param>
		/// <returns>An enumeratio</returns>
		public IList SelectAll (Type objectType , Type collectionType ) 
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			return data.SelectAll( collectionType );
		}


		/// <summary>
		/// Gets all the objects of a certain type form the database based of a sql filter
		/// </summary>
		/// <param name="objectType">Type of obejct to get</param>
		/// <param name="filter">The filter to apply</param>
		/// <param name="orderby">Sort string</param>
		/// <returns>The collection of objects</returns>
		public IList SelectBy (Type objectType, WhereClause wherePhrase, OrderByItemCollection orderByItems) 
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			return data.SelectBy(wherePhrase, orderByItems);
		}

        public IList SelectBy(Type objectType, string wherePhrase, string orderBy)
        {
            ObjectDataAccess data = GetObjectDataAccess(objectType);
            return data.SelectBy(wherePhrase, orderBy);
        }

        public List<T> SelectBy<T>(WhereClause wherePhrase, OrderByItemCollection orderByItems)
        {
            ObjectDataAccess data = GetObjectDataAccess(typeof(T));
            return (List<T>) GetGenericListFromObjectList<T> (data.SelectBy (wherePhrase, orderByItems));
        }

        public List<T> SelectBy<T>(string whereClause, string orderBy)
        {
            ObjectDataAccess data = GetObjectDataAccess(typeof(T));
            return (List<T>)GetGenericListFromObjectList<T>(data.SelectBy(whereClause, orderBy));
        }



		/// <summary>
		/// Gets top n number of a certain type form the database based of a sql filter
		/// </summary>
		/// <param name="objectType">Type of obejct to get</param>
		/// <param name="top">Number of items to get from the database</param>
		/// <param name="filter">The filter to apply</param>
		/// <param name="orderby">Sort string</param>
		/// <returns>The collection of objects</returns>
        public IList SelectTopBy(Type objectType, int top, WhereClause wherePhrase, OrderByItemCollection orderByItems) 
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			return data.SelectTopBy (top, wherePhrase, orderByItems);
		}

		/// <summary>
		/// Gets top n number of a certain type form the database based of a sql filter
		/// </summary>
		/// <param name="objectType">Type of obejct to get</param>
		/// <param name="top">Number of items to get from the database</param>
		/// <param name="orderby">Sort string</param>
		/// <returns>The collection of objects</returns>
		public IList SelectTopBy (Type objectType,  int top , OrderByItemCollection orderByItems) 
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			return data.SelectTopBy (top, orderByItems);
		}

		/// <summary>
		/// Executes the sql and from the sql object tries to extarct the 
		/// type specified
		/// </summary>
		/// <param name="objectType">Type of obejct to extract</param>
		/// <param name="sql">The sql to execute</param>
		/// <returns>Collection of objects</returns>
		public IList SelectBySQL (Type objectType,  string sql) 
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			return data.SelectBySQL (sql);
		}

        /// <summary>
        /// Executes the sql and from the sql object tries to extarct the 
        /// type specified
        /// </summary>
        /// <param name="objectType">Type of obejct to extract</param>
        /// <param name="sql">The sql to execute</param>
        /// <returns>Collection of objects</returns>
        public IList<T> SelectBySQL<T>(string sql)
        {
            ObjectDataAccess data = GetObjectDataAccess(typeof(T));
            IList list = data.SelectBySQL(sql);
            List<T> retList = new List<T>();
            foreach (object obj in list)
            {
                retList.Add((T)obj);
            }
            return retList;
        }


		/// <summary>
		/// Executes the sql and from the sql object tries to extarct the 
		/// type specified
		/// </summary>
		/// <param name="objectType">Type of obejct to extract</param>
		/// <param name="sql">The sql to execute</param>
		/// <param name="mappings">Property to column mappings to use</param>
		/// <returns>Collection of objects</returns>
		public IList SelectBySQL (Type objectType,  string sql, PropertyMaps mappings) 
		{
            ObjectDataAccess data = GetObjectDataAccess(objectType);
			return data.SelectBySQL (sql, mappings);
		}

	
		/// <summary>
		/// When given a IDataReader tries to extract object out of the reader
		/// </summary>
		/// <param name="objectType">Type of object to extract</param>
		/// <param name="reader">The reader</param>
		/// <returns>the collection of objects</returns>
		public IList SelectFromReader (Type objectType,  IDataReader reader)
		{
			ObjectDataAccess data = GetObjectDataAccess (objectType);
			/// TODO : Impelement a solid version of SelectFromReader method. Current one is fragile
			throw new NotImplementedException("Function not implementted yet!");
		}

		#endregion 

		#region "Utility Functions"
		/// <summary>
		/// Checks if there is a existing table storage ready for the specified type
		/// </summary>
		/// <param name="type">Type of obejct to check for storage</param>
		/// <returns>True if stoarage is ready</returns>
		public bool IsStorageReady(Type type)
		{
			const string TABLE_READY_QUERY = "select count(*) from dbo.sysobjects where id = object_id(N'{0}') and OBJECTPROPERTY(id, N'IsUserTable') = 1";
			string tableName = type.Name;
			TableMapAttribute tmap = CommonHelper.GetAttribute(
				type, typeof (TableMapAttribute)) as TableMapAttribute;
			if (tmap!=null)
			{
				tableName = type.Name;
			}
			string sql = string.Format(TABLE_READY_QUERY,tableName);
			DatabaseConnection dbConn = this.GetGraph().Connection;
			IDbConnection conn = dbConn.Open();
			IDbCommand cmd = DBObjectsFactory.CreateCommand(_configuration.Provider, sql,conn, dbConn.Transaction);
			int count = (int) cmd.ExecuteScalar() ;
			cmd.Connection = null;
			cmd.Dispose();
			dbConn.Close();
			if (count>0)
			{
				return true;
			}
			else
			{
				return false;
			}
		}

		/// <summary>
		/// Creates a table for the specified type
		/// </summary>
		/// <param name="type">Type of obejct for which storage need to be created</param>
		public void CreateStorage(Type type)
		{
			TableGenerator generator = new TableGenerator(_configuration.Provider);
			string sql = generator.GenerateTableScript(type);
			DatabaseConnection dbConn = this.GetGraph().Connection;
			IDbConnection conn = dbConn.Open();
			IDbCommand cmd = DBObjectsFactory.CreateCommand( _configuration.Provider, sql,conn, dbConn.Transaction);
			cmd.ExecuteNonQuery();
			cmd.Connection = null;
			cmd.Dispose();
			dbConn.Close();
		}

		/// <summary>
		/// Destroys the storage for the specified type
		/// </summary>
		/// <param name="type">Type of object for which storage is to br created</param>
		public void DestroyStorage (Type type)
		{
            DestroyStorage(type, false);
		}


        /// <summary>
        /// Destroys the storage for the specified type
        /// </summary>
        /// <param name="type">Type of object for which storage is to br created</param>
        public void DestroyStorage(Type type, bool checkExists)
        {
            bool proceed = true;
            if (checkExists)
            {
                proceed = IsStorageReady(type);
            }
            if (proceed )
            {
                TableGenerator generator = new TableGenerator(_configuration.Provider);
                string sql = generator.GenerateTableDropScript(type);
                DatabaseConnection dbConn = this.GetGraph().Connection;
                IDbConnection conn = dbConn.Open();
                IDbCommand cmd = DBObjectsFactory.CreateCommand(_configuration.Provider, sql, conn, dbConn.Transaction);
                cmd.ExecuteNonQuery();
                cmd.Connection = null;
                cmd.Dispose();
                dbConn.Close();
            }
        }

		/// <summary>
		/// Ensures storage for an object. If there is not a table in the database
		/// for the specified type it creates a table.
		/// </summary>
		/// <param name="type"></param>
		public void EnsureStorage(Type type)
		{
			if (!IsStorageReady(type))
			{
				CreateStorage(type);
			}
		}

		/// <summary>
		/// Executes non standard sql queries
		/// </summary>
		/// <param name="sql"></param>
		public void ExecuteNonQuery (string sql)
		{
			DatabaseConnection dbConn = this.GetGraph().Connection;
			IDbConnection conn = dbConn.Open();
			IDbCommand cmd = DBObjectsFactory.CreateCommand(_configuration.Provider, sql,conn, dbConn.Transaction);
			cmd.ExecuteNonQuery();
			cmd.Connection = null;
			cmd.Dispose();
			dbConn.Close();
		}

        ///// <summary>
        ///// Executes non standard sql queries
        ///// </summary>
        ///// <param name="sql"></param>
        //public DataSet ExecuteDataSet(string sql)
        //{	
        //    DatabaseConnection dbConn = this.GetGraph().Connection;
        //    IDbConnection conn = dbConn.Open();
        //    SqlCommand cmd = DBObjectsFactory.CreateCommand(sql,conn, dbConn.Transaction) as SqlCommand; 
        //    DataSet ds = new DataSet();
        //    SqlDataAdapter adapter = new SqlDataAdapter (cmd as SqlCommand);
        //    adapter.Fill (ds);
        //    cmd.Connection = null;
        //    cmd.Dispose();
        //    dbConn.Close();
        //    return ds;
        //}


		/// <summary>
		/// Fills up all related obejcts
		/// </summary>
		/// <param name="obj"></param>
		public void FillObjectRelations (object obj)
		{
			throw new NotSupportedException ("Not yet fully supported");
			ObjectDataAccess data = GetObjectDataAccess (obj.GetType());
			PropertyInfo[] pinfos = obj.GetType().GetProperties();
			
			foreach (PropertyInfo pinfo in pinfos)
			{
				ParentEntityAttribute pe =  CommonHelper.GetAttribute (pinfo, typeof(ParentEntityAttribute)) as ParentEntityAttribute;
				if (pe!=null)
				{
					// Need to get the property
					string propertyName = pe.ForeignKeyPropertyName;
					PropertyMap pmap = data.Schema.MappingByProperty [propertyName];
					object pk = pmap.Property.GetValue (obj, (object[])null);
					//object parentObject = this.Select(pinfo.PropertyType, pk);
					//pinfo.SetValue (obj,parentObject,(object[])null);
				}
			}
		}

		/// <summary>
		/// Registers an object type with the database schema. 
		/// </summary>
		/// <param name="objectType">Type of object</param>
		/// <returns>True or false indicating success or failure</returns>
		/// <example>
		///		The following code preloads the schema and maps it to object.
		///		<code>
		///			string connectionString ="Server=localhost;Database=ObjectSpace;Trusted_Connection=True;";
		///			// Creates a new database
		///			Database database = new Database(connectionString);
		///			// Preload the schema before call
		///			database.RegisterType (typeof(User));
		///			User user = new User();
		///			user.Name = "abcd";
		///			// If we did not preload the schema then it would be done inside the call
		///			database.Insert (user);
		///		</code>
		/// </example>
        public bool RegisterType(Type objectType)
		{
            return RegisterType (objectType, _configuration);
		}

		/// <summary>
		/// Registers an object type with the database schema. Internal
		/// </summary>
		/// <param name="objectType"></param>
		/// <param name="connectionString"></param>
		/// <returns></returns>
		/// <remarks>No documentation intended</remarks>
		private bool RegisterType (Type objectType, Configuration configuration)
		{
			lock (ObjectDataAccessDictionary)
			{
                if (IsTypeRegistered(objectType))
				{
					// Object already registered. Cannot register again
					return false;
				}
				else
				{
                    ObjectDataAccess data = new ObjectDataAccess(objectType, this.GetGraph().Connection, _configuration );
                    ObjectDataAccessDictionary.Add(objectType, data);
					return true;
				}
			}
		}

		/// <summary>
		/// Registers an object type with the database schema. Internal
		/// </summary>
		/// <param name="objectType"></param>
		/// <param name="data"></param>
		/// <returns></returns>
		/// <remarks>No documentation intended. Not used now</remarks>
		/// TODO: Comment required
		public bool RegisterType (Type objectType, ObjectDataAccess data )
		{
			lock (ObjectDataAccessDictionary)
			{
                if (IsTypeRegistered(objectType))
				{
					// Object already registered. Cannot register again
					return false;
				}
				else
				{
                    ObjectDataAccessDictionary.Add(objectType, data);
					return true;
				}
			}
		}


		/// <summary>
		/// Unregisters an object type. Use this to remove an existing schema map
		/// </summary>
		/// <param name="objectType">Type of object</param>
		/// <returns>True if the object was successfully unregistered</returns>
        public bool UnRegisterType(Type objectType)
		{
            if (!IsTypeRegistered(objectType))
			{
				// Object not registered. Cannot unregister 
				return false;
			}
			else
			{
                ObjectDataAccessDictionary.Remove(objectType);
				return true;
			}
		}

		/// <summary>
		/// Checks if the type's schema is registered with the database
		/// </summary>
		/// <param name="objectType">Type of object</param>
		/// <returns></returns>
		public bool IsTypeRegistered (Type objectType)
		{
			return ObjectDataAccessDictionary.ContainsKey (objectType);
		}

	
		/// <summary>
		/// Gets the objects data for a certain typr
		/// </summary>
		/// <param name="objectType"></param>
		/// <returns></returns>
		public ObjectDataAccess GetObjectDataAccess (Type objectType)
		{
            if (IsTypeRegistered(objectType))
			{
                return (ObjectDataAccess)ObjectDataAccessDictionary[objectType];
			}
			else
			{
                RegisterType(objectType);
                return (ObjectDataAccess)ObjectDataAccessDictionary[objectType];
			}
		}

		#endregion		
				
		#region "Private Methods"

        private List<T> GetGenericListFromObjectList<T>(IList list)
        {
            List<T> retList = new List<T>();
            foreach (object obj in list)
            {
                T safeObject = (T) obj ;
                retList.Add(safeObject);
            }
            return retList;
        }
        
        /// <summary>
		/// Contains schema for object types
		/// </summary>
		private Hashtable ObjectDataAccessDictionary
		{
			get
			{
				return this.GetGraph().ObjectDataDictionary;
			}
		}


		private DatabaseGraph GetGraph ()
		{
			if (!_pooled)
			{
				if (_graph==null)
				{
					_graph = new DatabaseGraph (_configuration);
				}
				return _graph;
			} 
			else
			{
				return Database.GetGraph(_configuration);
			}
		}


		public static void PurgeCache (bool alsoFlushLiveObjects)
		{
			if (alsoFlushLiveObjects)
			{
				lock (_htDatabaseGraphs)
				{
					_htDatabaseGraphs.Clear();
				}
			}
			else
			{
				lock (_htDatabaseGraphs)
				{
					Hashtable htThreads = new Hashtable();
					ProcessThreadCollection threads = Process.GetCurrentProcess().Threads;
					foreach (ProcessThread  thread in threads)
					{
						htThreads.Add (thread.Id,thread.Id);
					}

					ArrayList listDead = new ArrayList();
					foreach (object key in _htDatabaseGraphs.Keys)
					{
						DatabaseGraph graph = (DatabaseGraph) _htDatabaseGraphs[key];
						int threadID = graph.ThreadID;
						if (!htThreads.ContainsKey (threadID))
						{
							listDead.Add (key);
						}
					}

					foreach (object key in listDead)
					{
						_htDatabaseGraphs.Remove (key);
					}
				}
			}
		}

		#endregion 

		#region "Static methods and variables"

		// The dictionary of conection strings and databases
		private static Hashtable _htDatabaseGraphs = new Hashtable();
		/// <summary>
		/// Gets the database graph for a spcified connection
		/// </summary>
		/// <param name="connectionString">Connection string</param>
		/// <returns>The database graph</returns>
		private static DatabaseGraph GetGraph (Configuration configuration)
		{
			string hashKey="";
            if (configuration.ConnectionString.ToLower().IndexOf("trusted_connection=true") != -1)
			{
				WindowsPrincipal wp = new WindowsPrincipal(WindowsIdentity.GetCurrent());
				string username = "" + wp.Identity.Name;

                hashKey = configuration.ConnectionString + "#" + AppDomain.GetCurrentThreadId().ToString() 
					+ "#" + username;
			}
			else
			{
                hashKey = configuration.ConnectionString + "#" + AppDomain.GetCurrentThreadId().ToString();
			}

			if (_htDatabaseGraphs.ContainsKey (hashKey))
			{
				return (DatabaseGraph) _htDatabaseGraphs[hashKey];
			}
			else
			{
				lock (_htDatabaseGraphs)
				{
					DatabaseGraph graph = new DatabaseGraph (configuration);
					_htDatabaseGraphs.Add (hashKey, graph);
				}
				return (DatabaseGraph) _htDatabaseGraphs[hashKey];
			}
		}
		#endregion

		#region IDisposable Members

		public void Dispose()
		{
			if( _graph != null )
				_graph.Dispose();
		}

		#endregion
	}
}
